#include "chunk.h"
#include "malloc.h"
#include "chunk_manager.h"
#include "block.h"
#include "raw_chunk_manager.h"

#include <unistd.h>
#include <stdbool.h>

void free_tiny(chunk_t *chunk)
{
	chunk_t chunk_prev, chunk_next, *first_chunk = chunk, *last_chunk = chunk;

	if (chunk->prev)
	{
		chunk_read(chunk->prev, &chunk_prev);
		if (chunk_prev.is_used == false && chunk_prev.block_id == chunk->block_id)
			first_chunk = &chunk_prev;
	}
	if (chunk->next)
	{
		chunk_read(chunk->next, &chunk_next);
		if (chunk_next.is_used == false && chunk_next.block_id == chunk->block_id)
			last_chunk = &chunk_next;
	}

	first_chunk->is_used = false;
	chunk_merge(first_chunk, last_chunk);

	if (is_alone_on_block(first_chunk))
	{
		if (first_chunk->current == allocs_tree[TINY])
			allocs_tree[TINY] = last_chunk->next;
		destroy_block(first_chunk);
	}
}

void free_small(chunk_t *chunk)
{
	(void) chunk;
}

void free_large(chunk_t *chunk)
{
	if (chunk->current == allocs_tree[LARGE])
		allocs_tree[LARGE] = chunk->next;
	destroy_block(chunk);
}

static void (*free_funcs[3])(chunk_t*) = {free_tiny, free_small, free_large};

void ft_free(void *ptr)
{
	chunk_t chunk;
    int category;

    if (get_chunk(ptr, &chunk, &category))
    {
#ifdef DEBUG
        write(2, "chunk not found\n", 16);
#endif
        return;
    }
#ifdef DEBUG
	if (chunk.is_used == false)
		write(2, "double free\n", 12);
#endif
	// CALL THE RIGHT FREE FUNCTION DEPEND ON SIZE
	free_funcs[category](&chunk);	
}